# hobo - portal to configure and deploy applications
# Copyright (C) 2015-2022 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import urllib
import urllib.parse

from django.conf import settings
from requests import Session as RequestsSession
from requests.auth import AuthBase

from hobo.signature import sign_url


class PublikSignature(AuthBase):
    def __init__(self, orig, secret):
        self.orig = orig
        self.secret = secret

    def __call__(self, request):
        url = request.url

        parsed_url = urllib.parse.urlparse(url)
        query_params = urllib.parse.parse_qs(parsed_url.query)
        query_params['orig'] = [self.orig]
        query = urllib.parse.urlencode(query_params, doseq=True)
        url = urllib.parse.urlunparse(parsed_url._replace(query=query))

        request.url = sign_url(url, self.secret)
        return request


def get_known_service_for_url(url):
    netloc = urllib.parse.urlparse(url).netloc
    scheme = urllib.parse.urlparse(url).scheme

    for services in settings.KNOWN_SERVICES.values():
        for service in services.values():
            remote_url = service.get('url')
            assert remote_url is not None

            remote_netloc = urllib.parse.urlparse(remote_url).netloc
            remote_scheme = urllib.parse.urlparse(remote_url).scheme

            if remote_netloc == netloc and remote_scheme == scheme:
                return service
    return None


class ImplicitPublikSignature(AuthBase):
    def __call__(self, request):
        url = request.url

        remote_service = get_known_service_for_url(url)
        if (
            remote_service
            and (secret := remote_service.get('secret'))
            and (orig := remote_service.get('orig'))
        ):
            return PublikSignature(orig=orig, secret=secret)(request)

        return request


class Requests(RequestsSession):
    def request(self, method, url, **kwargs):
        if not kwargs.get('auth'):
            kwargs['auth'] = ImplicitPublikSignature()
        return super().request(method, url, **kwargs)
